"""Model classes"""

from julesTk import ThreadSafeObject
from julesTk.utils.observe import Observable

__author__ = "Joeri Jongbloets <joeri@jongbloets.net>"


class Model(Observable, ThreadSafeObject):

    def __init__(self):
        super(Model, self).__init__()
        self._data = None

    @property
    def data(self):
        """RAW Representation of the data contained in the model"""
        with self.lock:
            result = self._data
        return result

    def update(self):
        """Request the model to update it self"""
        raise NotImplementedError



"""Implementing basic view objects"""

from julesTk import *

__author__ = "Joeri Jongbloets <joeri@jongbloets.net>"


class Frame(ttk.Frame, object):

    FONT_FAMILY = "Verdana"
    FONT_SMALL = (FONT_FAMILY, 8)
    FONT_NORMAL = (FONT_FAMILY, 10)
    FONT_LARGE = (FONT_FAMILY, 14)

    @staticmethod
    def configure_grid(w, sticky="nsew", **kwargs):
        w.grid(sticky=sticky, **kwargs)

    @staticmethod
    def configure_column(w, indexes, weight=1, **kwargs):
        if not isinstance(indexes, (tuple, list)):
            indexes = [indexes]
        kwargs["weight"] = weight
        for index in indexes:
            w.grid_columnconfigure(index, **kwargs)

    @staticmethod
    def configure_row(w, indexes, weight=1, **kwargs):
        if not isinstance(indexes, (tuple, list)):
            indexes = [indexes]
        kwargs["weight"] = weight
        for index in indexes:
            w.grid_rowconfigure(index, **kwargs)

    def __init__(self, parent):
        """

        :rtype: Tkinter.Frame
        """
        ttk.Frame.__init__(self, parent)
        self._parent = parent

    @property
    def parent(self):
        return self._parent

    def setup(self):
        raise NotImplementedError


class View(Frame):
    """Interacting frame"""

    def __init__(self, parent, controller):
        super(View, self).__init__(parent)
        self._controller = controller
        self._variables = {}
        self._widgets = {}

    @property
    def controller(self):
        """ The controller

        :rtype: julesTk.controller.Controller
        """
        return self._controller

    @property
    def variables(self):
        """Dictionary of variables registered to this view

        :rtype: dict[str, tk.Variable]
        """
        return self._variables

    def has_variable(self, name):
        """Returns whether a variable is registered to this view using the given name

        :rtype: bool
        """
        return name in self.variables.keys()

    def get_variable(self, name):
        """Returns a variable registered to this view

        :param name: Name of the variable
        :type name: str
        :return: The variable
        :rtype: tk.Variable
        """
        if not self.has_variable(name):
            raise KeyError("Unknown variable: {}".format(name))
        return self.variables[name]

    def add_variable(self, name, variable):
        """Registers a variable under a new name to this view

        :param name: The name under which the variable should be registered
        :type name: str
        :param variable: The variable to register
        :type variable: tk.Variable
        :return: The registered variable
        :rtype: tk.Variable
        """
        if self.has_variable(name):
            raise KeyError("Already registered a variable as: {}".format(name))
        self.variables[name] = variable
        return variable

    def remove_variable(self, name):
        """Removes a variable (name) from the registry

        :return: Whether the variable is still in the registry
        :rtype: bool
        """
        if not self.has_variable(name):
            raise KeyError("Unknown variable: {}".format(name))
        self.variables.pop(name)
        return not self.has_variable(name)

    @property
    def widgets(self):
        """Dictionary of widgets registered to this view

        :rtype: dict[str, tk.Widget]
        """
        return self._widgets

    def has_widget(self, name):
        """Whether this view has a widget registered under the given name

        :param name: Name to look for
        :type name: str
        :rtype: bool
        """
        return name in self.widgets.keys()

    def get_widget(self, name):
        """Returns the widget registered to this view under the given name

        :param name: Name to look for
        :type name: str
        :rtype: tk.Widget
        """
        if not self.has_widget(name):
            raise KeyError("Unknown widget: {}".format(name))
        return self.widgets[name]

    def add_widget(self, name, widget):
        """Registers a widget under a new name to the view

        :raises: KeyError, if name is already used

        :param name: Name to register the widget under
        :type name: str
        :param widget: The widget to register
        :type widget: tk.Widget
        :return: The widget that as registered
        :rtype: tk.Widget
        """
        if self.has_widget(name):
            raise KeyError("Already registered a widget as: {}".format(name))
        self.widgets[name] = widget
        return widget

    def remove_widget(self, name):
        """Removes a widget (name) from the registry

        :param name: Name to remove
        :type name: str
        :return: Whether the widget is still registered
        :rtype: bool
        """
        if not self.has_widget(name):
            raise KeyError("Unknown widget: {}".format(name))
        self.widgets.pop(name)
        return name not in self.widgets.keys()

    def show(self):
        """Shows the view"""
        self.tkraise()

    def hide(self):
        """Hides the view (temporarily)"""
        self.grid_remove()

    def close(self):
        """Closes the view"""
        return True

    def setup(self):
        """Configure this view"""
        raise NotImplementedError


class ViewSet(View):
    """A view can contain one or more views"""

    def __init__(self, parent, controller):
        super(ViewSet, self).__init__(parent, controller)
        self._views = {}

    @property
    def views(self):
        """

        :rtype: dict[str, pyLabJackView.view.View]
        """
        return self._views

    def get_view(self, name):
        """

        :param name: Name of the view
        :rtype: julesTk.view.View
        """
        if not self.has_view(name):
            KeyError("No view registered under: {}".format(name))
        return self.views[name]

    def has_view(self, name):
        return name in self.views

    def add_view(self, name, frame):
        if self.has_view(name):
            KeyError("Already registered a view under: {}".format(name))
        self.views[name] = frame

    def remove_view(self, name):
        if not self.has_view(name):
            KeyError("No view registered under: {}".format(name))
        self.get_view(name).close()
        self.views.pop(name)

    def show_view(self, name):
        """Show a specific View"""
        v = self.get_view(name)
        v.show()

    def hide_views(self):
        """Hide all Views"""
        for v in self._views.keys():
            self.hide_view(v)
        # done

    def hide_view(self, name):
        """Hide a specific View"""
        v = self.get_view(name)
        v.grid_remove()

    def show(self):
        """Show this ViewSet"""
        raise NotImplementedError

    def hide(self):
        """Hide this ViewSet"""
        raise NotImplementedError

    def setup(self):
        """Configure this ViewSet"""
        raise NotImplementedError

    def close(self):
        """Close this ViewSet"""
        while len(self.views.values()) > 0:
            v = self.views.keys()[0]
            self.get_view(v).close()
            self.remove_view(v)
        super(ViewSet, self).close()



"""Provides classes that help with creating Observer-Observable structures"""

import functools

__author__ = "Joeri Jongbloets <joeri@jongbloets.net>"


class Observable(object):
    """An Observable will update it's observers whenever it is changed

    Notification can be triggered in two ways:

    1. Call `notify_observers()` directly from a method, after data was updated
    2. Use `@observed` decorator for methods that should call `notify_observers` when finished

    Observers can register themselves using the `register_observer` method.
    Only 1 observer registration per object.

    """

    def __init__(self):
        # super prevents disruption of mro chain
        super(Observable, self).__init__()
        self._observers = []

    def register_observer(self, observer):
        """Register an observer to get notified when this object is changed.

        Will only add the observer if it is not already observing.

        :param observer: The observer that should be notified
        :type observer: julesTk.Observer
        """
        if not isinstance(observer, Observer):
            raise ValueError("Expected a Observer, not {}".format(type(observer)))
        if observer not in self._observers:
            self._observers.append(observer)

    def notify_observers(self):
        """Notifies all observing observers"""
        for observer in self._observers:
            observer.update(self)

    def observed(f):
        """Decorator that will automatically call notify_observers after executing this method

        :type f: callable
        """
        @functools.wraps(f)
        def magic(self, *args, **kwargs):
            result = f(self, *args, **kwargs)
            self.notify_observers()
            return result
        return magic

    observed = staticmethod(observed)


class Observer(object):
    """An observer for observing a observable

    Implement `update` to handle notifications
    """

    def __init__(self):
        super(Observer, self).__init__()

    def update(self, observable):
        """Handle a update notification from an object that is being observed by this object.

        :param observable: Object that sent the notification
        :type observable: julesTk.Observable
        """
        raise NotImplementedError




from __future__ import absolute_import
from . import Model
import random


class RandomModel(Model):

    def __init__(self, mean=0, std=1):
        super(RandomModel, self).__init__()
        self._mean = mean
        self._std = std
        self.reset()

    @property
    def mean(self):
        with self.lock:
            result = self._mean
        return result

    @property
    def std(self):
        with self.lock:
            result = self._std
        return result

    @Model.observed
    def reset(self):
        with self.lock:
            self._data = []

    def generate(self):
        return random.gauss(self.mean, self.std)

    @Model.observed
    def update(self):
        with self.lock:
            self._data.append(self.generate())



"""Provide Controller classes"""

from julesTk.model import Model
from julesTk.view import View, tk

__author__ = "Joeri Jongbloets <joeri@jongbloets.net>"


class BaseController(object):
    """A bare-bones controller, without view or model"""

    def __init__(self, parent):
        super(BaseController, self).__init__()
        self._parent = parent

    def __del__(self):
        self.stop()

    def setup(self):
        """Configures the controller (set-up model, view)"""
        raise NotImplementedError

    def start(self):
        raise NotImplementedError

    def stop(self):
        raise NotImplementedError

    @property
    def parent(self):
        """Return reference to the parent object

        :rtype: julesTk.app.Application
        """
        return self._parent

    @property
    def application(self):
        result = self.parent
        if not isinstance(result, tk.Tk):
            result = self.parent.parent
        return result


class ViewController(BaseController):
    """A controller with a view (no model)"""

    VIEW_CLASS = View

    def __init__(self, parent, view=None):
        super(ViewController, self).__init__(parent=parent)
        self._view = view

    @property
    def view(self):
        """Return the view managed by this controller

        :rtype: julesTk.view.View
        """
        return self._view

    def has_view(self):
        """Whether the controller has a view attached"""
        return self.view is not None

    def start(self):
        """Start the controller and open the view"""
        self.view.show()

    def stop(self):
        """Stop the controller and close the view managed by the view"""
        self.view.close()

    def setup(self):
        """Configure the controller and view.

        Should be called before showing the view or starting the controller.
        Preferably after initialization of the controller.

        :rtype: julesTk.controller.ViewController
        """
        if not self.has_view():
            if not issubclass(self.VIEW_CLASS, View):
                raise ValueError("Expected a view not {}".format(self.VIEW_CLASS.__name__))
            pview = self.parent
            if isinstance(pview, BaseController):
                pview = self.parent.view
            self._view = self.VIEW_CLASS(pview, self)
        self.view.setup()
        return self


class ModelController(BaseController):
    """A controller with a model (no view)"""

    def __init__(self, parent, model=None):
        super(ModelController, self).__init__(parent=parent)
        self._model = model

    @property
    def model(self):
        """The model used by this controller

        :return:
        :rtype: julesTk.model.Model
        """
        return self._model

    @model.setter
    def model(self, model):
        self._set_model(model)

    def _set_model(self, model):
        if not isinstance(model, Model):
            raise ValueError("Expected a model class")
        self._model = model

    def has_model(self):
        """Whether the controller has a model attached"""
        return self.model is not None

    def start(self):
        """Start the controller"""
        raise NotImplementedError

    def stop(self):
        """Stop the controller"""
        raise NotImplementedError

    def setup(self):
        """Set-up and configure the controller"""
        raise NotImplementedError


class Controller(ViewController, ModelController):
    """A controller with both a model and view"""

    def __init__(self, parent, view=None, model=None):
        ViewController.__init__(self, parent=parent, view=view)
        ModelController.__init__(self, parent=parent, model=model)



"""Module managing starting and stopping the application

"""

from . import tk, ThreadSafeObject

__author__ = "Joeri Jongbloets <joeri@jongbloets.net>"


class Application(ThreadSafeObject, tk.Tk):
    """Application entry point"""

    def __init__(self, *args, **kwargs):
        tk.Tk.__init__(self, *args, **kwargs)
        super(Application, self).__init__()
        self._controllers = {}

    @property
    def controllers(self):
        """All controllers attached to this application

        :return: A dictionary of the name - controller mapping
        :rtype: dict[str, pyLabJackView.controller.BaseController]
        """
        return self._controllers

    def get_controller(self, name):
        """Returns the controller registered under the given name

        :raises KeyError: If no controller is registered with that name

        :param name: Name of the controller
        :type name: str
        :rtype: julesTk.controller.BaseController
        """
        if not self.has_controller(name):
            raise KeyError("No controller registered under: {}".format(name))
        return self.controllers[name]

    def has_controller(self, name):
        """Returns whether a controller is registered under the given name

        :param name: Name of the controller
        :type name: str
        :rtype: bool
        """
        return name in self.controllers.keys()

    def add_controller(self, name, controller):
        """Registers a new controller under the given name

        :raises KeyError: If another controller is already registered under the same name

        :param name: Name to register the controller under
        :type name: str
        :param controller: The controller to register
        :type controller: julesTk.controller.BaseController
        """
        if self.has_controller(name):
            raise KeyError("Already registered a controller under: {}".format(name))
        self.controllers[name] = controller

    def remove_controller(self, name):
        """Removes a controller from the application

        :raises KeyError: If no controller is registered with that name
        :param name: Name of the controller
        :type name: str
        """
        if not self.has_controller(name):
            raise KeyError("No controller registered under: {}".format(name))
        self.controllers.pop(name)

    def setup(self):
        """Configures the application and loads at least one controller"""
        raise NotImplementedError

    def run(self):
        """Start the application"""
        with self.lock:
            # prepare the application
            self.setup()
            # now start the application
            self.start()
        # in the main loop we wait, so we do not need a lock
        self.mainloop()
        with self.lock:
            # clean up
            self.stop()

    def start(self):
        """Everything to show and run the applciation"""
        raise NotImplementedError

    def stop(self):
        while len(self.controllers) > 0:
            name = self.controllers.keys()[0]
            self.get_controller(name).stop()
            self.remove_controller(name)
        tk.Tk.quit(self)



"""Some basic classes"""

from threading import RLock

import sys
if sys.version_info[0] < 3:
    import Tkinter as tk
else:
    import tkinter as tk
import ttk
import functools

__author__ = "Joeri Jongbloets <joeri@jongbloets.net>"


class JulesException(BaseException):
    """A simple exception with the option to add messages"""

    def __init__(self, msg=None):
        if msg is None:
            self._msg = self.__class__.__name__
        self._msg = msg

    def __str__(self):
        return self._msg


class ThreadSafeObject(object):
    """A class providing infrastructure for creating thread-safe objects"""

    def __init__(self):
        super(ThreadSafeObject, self).__init__()
        self._lock = RLock()

    @property
    def lock(self):
        """Retrieve the lock of this object"""
        return self._lock

    def thread_safe(f):
        """A decorator for making methods thread-safe"""
        @functools.wraps(f)
        def magic(self, *args, **kwargs):
            try:
                self.lock.acquire()
                result = f(self, *args, **kwargs)
            finally:
                self.lock.release()
            return result
        return magic

    thread_safe = staticmethod(thread_safe)
